@@ -1,4 +1,4 @@
-__version__ = "0.0.6"
+__version__ = "0.0.7"
 
 from ._stack import UndoManager
 from ._const import empty
@@ -7,6 +7,7 @@ from . import abc, fmt, containers
 __all__ = [
     "UndoManager",
     "empty",
+    "arguments",
     "AbstractUndoableDict",
     "AbstractUndoableList",
     "AbstractUndoableSet",
@@ -31,3 +32,15 @@ def __getattr__(name):
         DeprecationWarning,
     )
     return cls
+
+
+def arguments(*args, **kwargs):
+    """
+    Function that makes returning arguments from a function easier.
+
+    Examples
+    --------
+    >>> arguments(1, 2)  # returns (1, 2), {}
+    >>> arguments(1, 2, a=3)  # returns (1, 2), {'a': 3}
+    """
+    return args, kwargs
