@@ -1,8 +1,10 @@
 from __future__ import annotations
 from abc import ABC, abstractmethod
-from typing import Any, Callable, Iterable, Iterator, TYPE_CHECKING
+from functools import lru_cache
+from typing import Any, Callable, Iterable, Iterator, TYPE_CHECKING, Mapping
+import inspect
 
-from ._reversible import ReversibleFunction
+from collections_undo._reversible import ReversibleFunction
 
 FormatterType = Callable[["Command"], str]
 
@@ -76,10 +78,19 @@ class Command(_CommandBase):
 
     @property
     def function_id(self) -> int:
+        """Return the function id."""
         return self.func.function_id
 
+    @lru_cache(maxsize=128)
+    def bind_args(self) -> inspect.BoundArguments:
+        return inspect.signature(self.func._func_fw).bind(*self.args, **self.kwargs)
+
     @classmethod
-    def merge(cls, cmds: Iterable[Command], formatter=None) -> CommandGroup:
+    def merge(
+        cls,
+        cmds: Iterable[Command],
+        formatter: FormatterType | None = None,
+    ) -> CommandGroup:
         group = CommandGroup(cmds, formatter=formatter)
         return group
 
@@ -100,6 +111,19 @@ class Command(_CommandBase):
             return self.func.format_reverse_call(*self.args, **self.kwargs)
         return self.func.format_forward_call(*self.args, **self.kwargs)
 
+    def reduce_with(self, cmd: Command) -> Self:
+        """Automatically merge the command with the given command."""
+        rule = cmd.func._reduce_rule
+        if rule is None:
+            raise ValueError(f"Automerge rule is not defined for {cmd.func}.")
+        if self.func is not cmd.func:
+            raise ValueError(f"Cannot merge different functions.")
+        _args, _kwargs = rule(
+            Arguments(self.bind_args().arguments),
+            Arguments(cmd.bind_args().arguments),
+        )
+        return self.__class__(self.func, _args, _kwargs)
+
 
 class CommandGroup(_CommandBase):
     """A group of commands."""
@@ -173,3 +197,36 @@ class CommandGroup(_CommandBase):
 
     def _format_default(self) -> str:
         return "\n".join(cmd.format() for cmd in self)
+
+
+class Arguments(Mapping[str, Any]):
+    def __init__(self, *args, **kwargs):
+        self._dict = dict(*args, **kwargs)
+        self._keys = list(self._dict.keys())
+
+    def __repr__(self):
+        d = repr(self._dict)
+        return f"Arguments({d[1:-1]})"
+
+    def __getitem__(self, key: str | int) -> Any:
+        if not isinstance(key, str):
+            key = self._keys[key]
+        return self._dict[key]
+
+    def __getattr__(self, key: str):
+        return self.__getitem__(key)
+
+    def __iter__(self):
+        return iter(self._keys)
+
+    def __len__(self):
+        return len(self._keys)
+
+    def keys(self):
+        return self._keys
+
+    def values(self):
+        return self._dict.values()
+
+    def items(self):
+        return self._dict.items()
