@@ -1,13 +1,9 @@
 from __future__ import annotations
 from contextlib import contextmanager
-from enum import Enum
 from typing import (
     Any,
     Callable,
-    Iterator,
     Literal,
-    MutableSequence,
-    NamedTuple,
     TYPE_CHECKING,
     TypeVar,
     overload,
@@ -18,76 +14,17 @@ from ._reversible import ReversibleFunction
 from ._undoable import UndoableInterface, UndoableProperty
 from ._command import Command, _CommandBase
 from ._const import empty
+from ._stack_utils import LengthPair, CallbackList, CallType
 
 if TYPE_CHECKING:
-    from typing_extensions import Self
+    from typing_extensions import Self, ParamSpec
+
+    _P = ParamSpec("_P")
+    _R = TypeVar("_R")
 
 _F = TypeVar("_F", bound=Callable)
 
 
-class CallbackList(MutableSequence[_F]):
-    """
-    A list of callbacks of UndoManager updates.
-
-    Callbacks are useful for such as logging.
-    """
-
-    def __init__(self) -> None:
-        self._list = []
-
-    def insert(self, index: int, callback: _F, /):
-        self._check_callable(callback)
-        self._list.insert(index, callback)
-
-    def __repr__(self) -> str:
-        return f"{type(self).__name__}({self._list!r})"
-
-    def __len__(self) -> int:
-        return len(self._list)
-
-    def __iter__(self) -> Iterator[_F]:
-        return iter(self._list)
-
-    def __getitem__(self, key):
-        return self._list[key]
-
-    def __setitem__(self, key, callback):
-        self._check_callable(callback)
-        self._list[key] = callback
-
-    def __delitem__(self, key):
-        del self._list[key]
-
-    def evoke(self, *args, **kwargs) -> None:
-        """Evoce all callbacks."""
-        for callback in self._list:
-            callback(*args, **kwargs)
-        return None
-
-    @staticmethod
-    def _check_callable(obj):
-        if not callable(obj):
-            raise TypeError("Can only insert callable object to the callback list.")
-
-
-class CallType(Enum):
-    call = "call"
-    undo = "undo"
-    redo = "redo"
-
-    def __eq__(self, other):
-        if isinstance(other, str):
-            return self.value == other
-        return super().__eq__(other)
-
-
-class LengthPair(NamedTuple):
-    """Pair of stack size."""
-
-    undo: int
-    redo: int
-
-
 def always_zero(*args, **kwargs) -> float:
     """Default command size counter."""
     return 0.0
@@ -102,6 +39,8 @@ class ManagerState:
         self.measure = measure
         self.maxsize = maxsize
         self.is_blocked = False
+        self.is_merging = False
+        self.is_reducing = False
         self.stack_undo: list[_CommandBase] = []
         self.stack_redo: list[_CommandBase] = []
         self.stack_undo_size = 0.0
@@ -243,7 +182,18 @@ class UndoManager:
         if self.is_blocked:
             return None
 
-        self._state.stack_undo.append(cmd)
+        if self._state.is_reducing and len(self._state.stack_undo) > 0:
+            last_cmd = self._state.stack_undo[-1]
+            if isinstance(last_cmd, Command):
+                new_cmd = last_cmd.reduce_with(cmd)
+                popped_cmd = self._state.stack_undo.pop(-1)
+                self._state.stack_undo_size -= popped_cmd.size
+            else:
+                new_cmd = cmd
+            self._state.stack_undo.append(new_cmd)
+        else:
+            self._state.stack_undo.append(cmd)
+
         self._state.stack_redo.clear()
         self.called.evoke(cmd, CallType.call)
 
@@ -270,7 +220,9 @@ class UndoManager:
         return None
 
     @overload
-    def undoable(self, f: Callable, name: str | None = None) -> ReversibleFunction:
+    def undoable(
+        self, f: Callable[_P, _R], name: str | None = None
+    ) -> ReversibleFunction[_P, _R, Any]:
         ...
 
     @overload
@@ -282,7 +234,7 @@ class UndoManager:
         self,
         f: Literal[None],
         name: str | None = None,
-    ) -> Callable[[Callable], ReversibleFunction] | Callable[
+    ) -> Callable[[Callable[_P, _R]], ReversibleFunction[_P, _R, Any]] | Callable[
         [property], UndoableProperty
     ]:
         ...
@@ -311,7 +263,9 @@ class UndoManager:
         return UndoableProperty(fget, fset, fdel, doc=doc, mgr=self)
 
     @overload
-    def interface(self, func: Callable, name: str | None = None) -> UndoableInterface:
+    def interface(
+        self, func: Callable[_P, _R], name: str | None = None
+    ) -> UndoableInterface[_P, _R, Any]:
         ...
 
     @overload
@@ -319,7 +273,7 @@ class UndoManager:
         self,
         func: Literal[None],
         name: str | None = None,
-    ) -> Callable[[Callable], UndoableInterface]:
+    ) -> Callable[[Callable[_P, _R]], UndoableInterface[_P, _R, Any]]:
         ...
 
     def interface(
@@ -363,13 +317,27 @@ class UndoManager:
     @contextmanager
     def merging(self, formatter: Callable | None = None) -> None:
         """Merge all the commands into a single command in this context."""
+        if self._state.is_merging:
+            yield None
+            return None
+
         blocked = self._state.is_blocked
+        merging = self._state.is_merging
         len_before = len(self._state.stack_undo)
-        yield None
-        if not blocked:
-            len_after = len(self._state.stack_undo)
-            self.merge_commands(len_before, len_after, formatter=formatter)
-            self.called.evoke(self._state.stack_undo[-1], CallType.call)
+        self._state.is_merging = True
+        try:
+            yield None
+        finally:
+            self._state.is_merging = merging
+            if not blocked and not merging:
+                len_after = len(self._state.stack_undo)
+                self.merge_commands(len_before, len_after, formatter=formatter)
+                self.called.evoke(self._state.stack_undo[-1], CallType.call)
+        return None
+
+    def set_merging(self, enabled: bool) -> None:
+        """Enable/disable merging."""
+        self._state.is_merging = bool(enabled)
         return None
 
     @contextmanager
@@ -392,6 +360,22 @@ class UndoManager:
             self.errored.evoke(e)
             raise e
 
+    @contextmanager
+    def reducing(self):
+        """Enable command reduction in this context."""
+        was_reducing = self._state.is_reducing
+        self._state.is_reducing = True
+        try:
+            yield None
+        finally:
+            self._state.is_reducing = was_reducing
+        return None
+
+    def set_reducing(self, enabled: bool):
+        """Enable/disable command reduction."""
+        self._state.is_reducing = bool(enabled)
+        return None
+
 
 def _join_stack(stack: list, max: int = 10):
     _splitter = ",\n    "
