@@ -1,17 +1,20 @@
 from __future__ import annotations
 from functools import partial, wraps
-from typing import Any, Callable, TYPE_CHECKING, Literal, TypeVar
-from ._reversible import ReversibleFunction
-from ._const import empty, FormatterType
+from typing import Any, Callable, TYPE_CHECKING, Generic, Literal, TypeVar
+from collections_undo._reversible import ReversibleFunction
+from collections_undo._const import empty, FormatterType, Args
 
 if TYPE_CHECKING:
     from typing_extensions import ParamSpec
-    from ._stack import UndoManager
+    from collections_undo._stack import UndoManager
 
     _P = ParamSpec("_P")
-    _R = TypeVar("_R")
-    ArgsType = tuple[tuple, dict[str, Any]]
-    _Args = TypeVar("_Args", bound=ArgsType)
+    _Args = TypeVar("_Args", bound=Args)
+else:
+    _P = TypeVar("_P")
+    _Args = TypeVar("_Args")
+
+_R = TypeVar("_R")
 
 
 def _dummy_func(*args, **kwargs) -> None:
@@ -21,7 +24,7 @@ def _dummy_func(*args, **kwargs) -> None:
 _Fmt = TypeVar("_Fmt", bound=FormatterType)
 
 
-class UndoableInterface:
+class UndoableInterface(Generic[_P, _R, _Args]):
     """
     An undoable object described by server/receiver interface.
 
@@ -52,7 +55,7 @@ class UndoableInterface:
         self._formatter_fw = None
         self._formatter_rv = None
 
-    def server(self, fserve: Callable[_P, _Args]) -> UndoableInterface:
+    def server(self, fserve: Callable[_P, _Args]) -> UndoableInterface[_P, _R, _Args]:
         """Set the server function."""
         itf = UndoableInterface(
             fserve=fserve,
@@ -63,7 +66,7 @@ class UndoableInterface:
         itf._formatter_rv = self._formatter_rv
         return itf
 
-    def receiver(self, freceive: Callable[_P, _R]) -> UndoableInterface:
+    def receiver(self, freceive: Callable[_P, _R]) -> UndoableInterface[_P, _R, _Args]:
         """Set the receiver function."""
         itf = UndoableInterface(
             fserve=self._fserve,
@@ -96,6 +99,7 @@ class UndoableInterface:
 
     @property
     def func(self) -> ReversibleFunction:
+        """Return the reversible function."""
         if self._func is None:
             self._func = self._create_function()
         return self._func
@@ -130,13 +134,17 @@ class UndoableInterface:
         return f"{type(self).__name__}<{self._freceive!r}>"
 
     def _create_function(self) -> ReversibleFunction:
-        def fw(new: ArgsType, old: ArgsType):
+        """Create a reversible function from the interface."""
+
+        # forward function
+        def fw(new: Args, old: Args):
             args, kwargs = new
             return self._freceive(*args, **kwargs)
 
         fw.__name__ = self.__name__
 
-        def rv(new: ArgsType, old: ArgsType):
+        # reverse function
+        def rv(new: Args, old: Args):
             if old is None:
                 return empty
             args, kwargs = old
@@ -151,8 +159,15 @@ class UndoableInterface:
             fn._formatter_rv = self._formatter_rv
 
         fn._map_args = _mapping
+        fn._reduce_rule = self._reduce_rule
         return fn
 
+    @staticmethod
+    def _reduce_rule(args0: dict[str, Any], args1: dict[str, Any]):
+        old = args0["old"]
+        new = args1["new"]
+        return (new, old), {}
+
 
 def _mapping(new, old):
     args, kwargs = new
@@ -205,6 +220,8 @@ class UndoableProperty(property):
             old_val = self.fget(obj)
             setattr.__get__(obj)(val, old_val)
 
+        setattr._reduce_rule = self._setter_reduce_rule
+
         # update names and the formatter
 
         return UndoableProperty(
@@ -248,3 +265,10 @@ class UndoableProperty(property):
 
     def _map_args(self, *args, **kwargs):
         return args, kwargs
+
+    @staticmethod
+    def _setter_reduce_rule(obj, args0: dict[str, Any], args1: dict[str, Any]):
+        # obj, val, old_val
+        old = args0["old_val"]
+        new = args1["val"]
+        return (new, old), {}
